
#include "sparsegraph.h"


SparseGraph *AbsDiff2SparseGraph(Image *img, float r){
  SparseGraph *sg=NULL;
  AdjRel *A=Circular(r);
  int p,q,i,n;
  int weight;
  int u_x,u_y,v_x,v_y;

  n  = img->ncols*img->nrows;
  sg = CreateSparseGraph(img->ncols, 
			 img->nrows, A);
  sg->type = DISSIMILARITY;
  sg->Wmax = 0;

  for(p=0; p<n; p++){
    u_x = p%img->ncols; //PixelX(img,p);
    u_y = p/img->ncols; //PixelY(img,p);

    (sg->n_link[p])[0] = NIL;
    for(i=1; i<A->n; i++){
      v_x = u_x + A->dx[i];
      v_y = u_y + A->dy[i];
      if(ValidPixel(img,v_x,v_y)){
	q = v_x + img->ncols*v_y;

	weight = abs(img->data[p]-img->data[q]);
	(sg->n_link[p])[i] = weight;

	if(weight>sg->Wmax) sg->Wmax = weight;
      }
      else{
	(sg->n_link[p])[i] = NIL;
      }
    }
  }
  return sg;
}


SparseGraph *WeightImage2SparseGraph(Image *W, float r){
  SparseGraph *sg=NULL;
  AdjRel *A=Circular(r);
  int p,q,i,n;
  int weight;
  int u_x,u_y,v_x,v_y;

  n  = W->ncols*W->nrows;
  sg = CreateSparseGraph(W->ncols, 
			 W->nrows, A);
  sg->type = DISSIMILARITY;
  sg->Wmax = 0;

  for(p=0; p<n; p++){
    u_x = p%W->ncols; //PixelX(W,p);
    u_y = p/W->ncols; //PixelY(W,p);

    (sg->n_link[p])[0] = NIL;
    for(i=1; i<A->n; i++){
      v_x = u_x + A->dx[i];
      v_y = u_y + A->dy[i];
      if(ValidPixel(W,v_x,v_y)){
	q = v_x + W->ncols*v_y;

	weight = W->data[p] + W->data[q];
	(sg->n_link[p])[i] = weight;

	if(weight>sg->Wmax) sg->Wmax = weight;
      }
      else{
	(sg->n_link[p])[i] = NIL;
      }
    }
  }
  return sg;
}


int get_edge_index(int p, int q, SparseGraph *g){
  AdjRel *A;
  int ncols,dx,dy,i;

  A = g->A;
  ncols = g->ncols;
  dx = q%ncols - p%ncols;
  dy = q/ncols - p/ncols;

  for(i=1; i<A->n; i++){
    if(dx == A->dx[i] && dy == A->dy[i])
      return i;
  }
  return -1;
}


SparseGraph   *CreateSparseGraph(int ncols, int nrows, AdjRel *A){
  SparseGraph *sg;
  int p,n,i;

  n = ncols*nrows;
  sg = (SparseGraph *)calloc(1, sizeof(SparseGraph));
  if(sg == NULL)
    Error(MSG1,"CreateSparseGraph");
  sg->A = A;
  sg->ncols = ncols;
  sg->nrows = nrows;
  sg->n_link = (int **)calloc(n,sizeof(int *));
  if(sg->n_link == NULL)
    Error(MSG1,"CreateSparseGraph");
  for(p=0; p<n; p++){
    sg->n_link[p] = AllocIntArray(A->n);
    for(i=0; i<A->n; i++)
      (sg->n_link[p])[i] = NIL;
  }
  sg->Wmax = 0;

  return sg;
}


void    DestroySparseGraph(SparseGraph **sg){
  SparseGraph *aux;  
  int p,n;

  aux = *sg;
  if(aux != NULL){
    n = aux->ncols*aux->nrows;
    DestroyAdjRel(&(aux->A));
    for(p=0; p<n; p++){
      free(aux->n_link[p]);
    }
    free(aux->n_link);
    free(*sg);
    *sg = NULL;
  }
}


SparseGraph   *CloneSparseGraph(SparseGraph *sg){
  SparseGraph *clone = NULL;
  int p,n,i;

  n = sg->ncols*sg->nrows;

  clone = CreateSparseGraph(sg->ncols, 
			    sg->nrows, 
			    CloneAdjRel(sg->A));
  for(p=0; p<n; p++){
    for(i=1; i<(sg->A)->n; i++){
      (clone->n_link[p])[i] = (sg->n_link[p])[i];
    }
  }
  clone->Wmax = sg->Wmax;
  clone->type = sg->type;

  return clone;
}



Image *ArcWeightImage(SparseGraph *sg){
  int i,p,n,weight;
  Image *arcw;
  AdjRel *A;

  A    = sg->A;
  arcw = CreateImage(sg->ncols,sg->nrows);
  n    = arcw->ncols*arcw->nrows;

  if(sg->type == CAPACITY){
    for(p=0; p<n; p++){
      arcw->data[p] = sg->Wmax;
      for(i=1; i<A->n; i++){
	weight = (sg->n_link[p])[i];
	if(weight==NIL) continue;
	arcw->data[p] = MIN(weight, arcw->data[p]);
      }
    }
  }
  else{ //DISSIMILARITY
    for(p=0; p<n; p++){
      arcw->data[p] = 0;
      for(i=1; i<A->n; i++){
	weight = (sg->n_link[p])[i];
	if(weight==NIL) continue;
	arcw->data[p] = MAX(weight, arcw->data[p]);
      }
    }
  }
  return(arcw);
}


void Orient2Digraph(SparseGraph *sg, Image *img, float per){
  AdjRel *A;
  int p,q,n,i,val,max = INT_MIN;
  int u_x,u_y,v_x,v_y;

  n = sg->ncols*sg->nrows;
  A = sg->A;

  for(p=0; p<n; p++){
    u_x = p%img->ncols;
    u_y = p/img->ncols;

    for(i=0; i<A->n; i++){
      if((sg->n_link[p])[i]==NIL) continue;
      v_x = u_x + A->dx[i];
      v_y = u_y + A->dy[i];
      if(!ValidPixel(img,v_x,v_y)) continue;
      q = v_x + img->ncols*v_y;

      val = (sg->n_link[p])[i];
      if(img->data[p] > img->data[q]){
	//val += ROUND(val*(per/100.0)); 
	val = ROUND(val*(1.0 + per/100.0));
      }
      else if(img->data[p] < img->data[q]){
	//val -= ROUND(val*(per/100.0)); 
	val = ROUND(val*(1.0 - per/100.0));
      }
      (sg->n_link[p])[i] = val;
      if(val > max) max = val;
    }
  }
  sg->Wmax = max;
}


void InvertArcsDigraph(SparseGraph *sg){
  AdjRel *A;
  int p,q,n,i,j,val_pq,val_qp;
  int u_x,u_y,v_x,v_y;

  n = sg->ncols*sg->nrows;
  A = sg->A;

  for(p=0; p<n; p++){
    u_x = p%sg->ncols;
    u_y = p/sg->ncols;

    for(i=0; i<A->n; i++){
      if((sg->n_link[p])[i]==NIL) continue;
      v_x = u_x + A->dx[i];
      v_y = u_y + A->dy[i];

      if(v_x < 0 || v_x >= sg->ncols ||
	 v_y < 0 || v_y >= sg->nrows)
	continue;

      q = v_x + sg->ncols*v_y;
      if(p > q){
	j = get_edge_index(q, p, sg);
	
	val_pq = (sg->n_link[p])[i];
	val_qp = (sg->n_link[q])[j];
	
	(sg->n_link[p])[i] = val_qp;
	(sg->n_link[q])[j] = val_pq;
      }
    }
  }
}



void Orient2DigraphInner(SparseGraph *sg, Image *P_sum){
  AdjRel *A;
  int p,q,n,i,val,max = INT_MIN;
  int u_x,u_y,v_x,v_y;

  n = sg->ncols*sg->nrows;
  A = sg->A;

  for(p=0; p<n; p++){
    u_x = p%sg->ncols;
    u_y = p/sg->ncols;

    for(i=1; i<A->n; i++){
      if((sg->n_link[p])[i]==NIL) continue;
      v_x = u_x + A->dx[i];
      v_y = u_y + A->dy[i];
      if(!ValidPixel(P_sum,v_x,v_y)) continue;
      q = v_x + sg->ncols*v_y;

      if (P_sum->data[q] == p) {
	sg->n_link[p][i] = 0;  //w(P_sum(q),q)=0
      }

      val = (sg->n_link[p])[i]; 

      if(val > max) max = val;
    }
  }
  sg->Wmax = max;
}

void Orient2DigraphOuter(SparseGraph *sg, Image *P_sum){
  AdjRel *A;
  int p,q,n,i,val,max = INT_MIN;
  int u_x,u_y,v_x,v_y;

  n = sg->ncols*sg->nrows;
  A = sg->A;

  for(p=0; p<n; p++){
    u_x = p%sg->ncols;
    u_y = p/sg->ncols;

    for(i=1; i<A->n; i++){
      if((sg->n_link[p])[i]==NIL) continue;
      v_x = u_x + A->dx[i];
      v_y = u_y + A->dy[i];
      if(!ValidPixel(P_sum,v_x,v_y)) continue;
      q = v_x + sg->ncols*v_y;

      if (P_sum->data[p] == q) {
	//j = get_edge_index(q, p, sg);
	sg->n_link[p][i] = 0;  // w(q,P_sum(q)) = 0
      }
      val = (sg->n_link[p])[i]; 

      if(val > max) max = val;
    }
  }
  sg->Wmax = max;
}


void  SuppressZeroWeightedArcs(SparseGraph *g){
  int i,p,n,wzero = 0;
  AdjRel *A;

  A = g->A;
  n = g->ncols*g->nrows;
  for(p=0; p<n && wzero==0; p++){
    for(i=1; i<A->n; i++){
      if((g->n_link[p])[i] == NIL)
	continue;

      if((g->n_link[p])[i] == 0){
	wzero = 1;
	break;
      }
    }
  }

  if(wzero==1){
    for(p=0; p<n; p++){
      for(i=1; i<A->n; i++){
	if((g->n_link[p])[i] == NIL)
	  continue;
	(g->n_link[p])[i] += 1; 
      }
    }
    g->Wmax++;
  }
}



void ChangeSparseGraphType(SparseGraph *sg, 
			   int type){
  int i,p,n;
  AdjRel *A;

  //printf("Change from %d to type: %d\n",sg->type, type);

  if(sg->type==type)
    return;

  A = sg->A;
  n = sg->ncols*sg->nrows;
  for(p=0; p<n; p++){
    //(sg->n_link[p])[0] = sg->Wmax or 0;
    for(i=1; i<A->n; i++){
      if((sg->n_link[p])[i]==NIL) continue;
      (sg->n_link[p])[i] = sg->Wmax - (sg->n_link[p])[i];
    }
  }
  if(sg->type==DISSIMILARITY)
    sg->type = CAPACITY;
  else
    sg->type = DISSIMILARITY;
}


// Increasing transformation.
void PowSparseGraph(SparseGraph *sg, 
		    int power, int max){
  AdjRel *A;
  int p,n,i;
  long double v,m;

  if(sg->Wmax<=1) return;
  if(power<=0)    return;

  n = sg->ncols*sg->nrows;
  A = sg->A;

  for(p=0; p<n; p++){
    for(i=0; i<A->n; i++){
      if((sg->n_link[p])[i]==NIL) continue;
      v = (double)((sg->n_link[p])[i]);
      m = powl((v/sg->Wmax),power);
      (sg->n_link[p])[i] = ROUND(m*(max-1))+1;
    }
  }
  sg->Wmax = max;
}



SparseGraph   *Convert2SymmetricSparseGraph(SparseGraph *g,
					    int method){
  SparseGraph *sg;
  AdjRel *A;
  int i,k,p,q,n,valpq,valqp,val=0;
  int ncols,nrows;
  int u_x,u_y,v_x,v_y;

  A     = g->A;
  ncols = g->ncols;
  nrows = g->nrows;
  n = ncols*nrows;

  sg = CreateSparseGraph(ncols, nrows,
			 CloneAdjRel(A));
  sg->type = g->type;
  sg->Wmax = 0;
  for(p=0; p<n; p++){
    u_x = p%ncols;
    u_y = p/ncols;
    for(i=1; i<A->n; i++){
      v_x = u_x + A->dx[i];
      v_y = u_y + A->dy[i];
      if(v_x>=0 && v_x<ncols && 
	 v_y>=0 && v_y<nrows){
	q = v_x + v_y*ncols;
	k = get_edge_index(q, p, g);

	valpq = (g->n_link[p])[i];
	valqp = (g->n_link[q])[k];
	switch(method){
	case 0:
	  val = (valpq + valqp)/2;
	  break;
	case 1:
	  val = MAX(valpq, valqp);
	  break;
	case 2:
	  val = MIN(valpq, valqp);
	  break;
	}
	(sg->n_link[p])[i] = val;
	(sg->n_link[q])[k] = val;
	if(val>sg->Wmax)
	  sg->Wmax = val;
      }
    }
  }

  return sg;
}




